@@ -3,38 +3,77 @@ import logging
 import os
 import scipy
 import flammkuchen
-import librosa
 import numpy as np
 from . import utils, data, models, event_utils, segment_utils, annot
-from typing import List, Optional, Dict, Any
+from typing import List, Optional, Dict, Any, Sequence
 import glob
+import tensorflow
+import zarr
+from tqdm.autonotebook import tqdm
+import dask.config
+import dask.array as da
+from dask.diagnostics import ProgressBar
+
+dask.config.set(**{'array.slicing.split_large_chunks': True})
 
 
-def predict_probabililties(x, model, params, verbose=None, prepend_data_padding: bool = True):
+def predict_probabilities(x: np.ndarray,
+                          model: tensorflow.keras.Model,
+                          params: Dict[str, Any],
+                          verbose: Optional[int] = 1,
+                          prepend_data_padding: bool = True):
     """[summary]
 
     Args:
         x ([samples, ...]): [description]
         model (tf.keras.Model): [description]
         params ([type]): [description]
-        verbose (int, optional): Verbose level for predict_generator (see tf.keras docs). Defaults to None.
+        verbose (int, optional): Verbose level for predict() (see keras docs). Defaults to 1.
         prepend_data_padding (bool, optional): Restores samples that are ignored
                     in the beginning of the first and the end of the last chunk
                     because of "ignore_boundaries". Defaults to True.
     Returns:
         y_pred - output of network for each sample [samples, nb_classes]
     """
+
     pred_gen = data.AudioSequence(x=x, y=None, shuffle=False, **params)  # prep data
-    y_pred = model.predict(pred_gen, verbose=verbose)  # run the network
-    y_pred = data.unpack_batches(y_pred, pred_gen.data_padding)  # reshape from [batches, nb_hist, ...] to [time, ...]
-    if prepend_data_padding:   # to account for loss of samples at the first and last chunks
-        y_pred = np.pad(y_pred,
-                        pad_width=((params['data_padding'], params['data_padding']), (0, 0)),
-                        mode='constant', constant_values=0)
-    return y_pred
+    nb_batches = len(pred_gen)
+    verbose = 1
+    prepend_data_padding = True
+
+    temp_store = zarr.storage.TempStore()
+    class_probabilities = zarr.zeros(shape=(0, params['nb_classes']),
+                                     chunks=(100_000, params['nb_classes']),
+                                     dtype='float',
+                                     store=temp_store,
+                                     overwrite=True)
+    # predict and unpack batch by batch to memmaped np or zarr array
+    for batch_number, batch_data in tqdm(enumerate(pred_gen), total=nb_batches, disable=verbose < 1):
+        y_pred_batch = model.predict_on_batch(batch_data)  # run the network
+
+        # reshape from [batches, nb_hist, ...] to [time, ...]
+        y_pred_unpacked_batch = data.unpack_batches(y_pred_batch, pred_gen.data_padding)
+
+        if prepend_data_padding:
+            pad_width = None
+            if batch_number == 0:  # to account for loss of samples at the first and last chunks
+                pad_width = ((params['data_padding'], 0), (0, 0))
+            elif batch_number == nb_batches - 1:
+                pad_width = ((0, params['data_padding']), (0, 0))
+
+            if pad_width is not None:
+                y_pred_unpacked_batch = np.pad(y_pred_unpacked_batch, pad_width=pad_width, mode='constant', constant_values=0)
+
+        class_probabilities.append(y_pred_unpacked_batch, axis=0)
+
+    class_probabilities = da.from_zarr(class_probabilities, inline_array=True)
+    class_probabilities = class_probabilities.rechunk((1_000_000, params['nb_classes']))
+    return class_probabilities
 
 
-def labels_from_probabilities(probabilities, threshold: Optional[float] = None):
+def labels_from_probabilities(probabilities,
+                              threshold: Optional[float] = None,
+                              indices: Optional[Sequence[int]] = None) -> np.ndarray:
     """Convert class-wise probabilities into labels.
 
     Args:
@@ -43,18 +82,23 @@ def labels_from_probabilities(probabilities, threshold: Optional[float] = None):
                                      If float, each class probability is compared to the threshold.
                                      First class to cross threshold wins.
                                      If no class crosses threshold label will default to the first class.
+        indices: (List[int], Optional): List of indices into axis 1 for which to compute the labels.
+                                     Defaults to None (use all indices).
     Returns:
         labels [samples,] - index of "winning" dimension for each sample
     """
+    if indices is None:
+        indices = slice(None)  # equivalent to ":"
+
     if probabilities.ndim == 1:
         if threshold is None:
             threshold = 0.5
-        labels = (probabilities > threshold).astype(np.intp)
+        labels = (probabilities[:, indices] > threshold).astype(np.intp)
     elif probabilities.ndim == 2:
         if threshold is None:
-            labels = np.argmax(probabilities, axis=1)
+            labels = np.argmax(probabilities[:, indices], axis=1)
         else:
-            thresholded_probabilities = probabilities.copy()
+            thresholded_probabilities = probabilities[:, indices].copy()
             thresholded_probabilities[thresholded_probabilities < threshold] = 0
             labels = np.argmax(thresholded_probabilities > threshold, axis=1)
     else:
@@ -65,8 +109,8 @@ def labels_from_probabilities(probabilities, threshold: Optional[float] = None):
 
 def predict_segments(class_probabilities: np.ndarray,
                      samplerate: float = 1.0,
-                     segment_dims: Optional[List[int]] = None,
-                     segment_names: Optional[List[str]] = None,
+                     segment_dims: Optional[Sequence[int]] = None,
+                     segment_names: Optional[Sequence[str]] = None,
                      segment_ref_onsets: Optional[List[float]] = None,
                      segment_ref_offsets: Optional[List[float]] = None,
                      segment_thres: float = 0.5,
@@ -107,69 +151,142 @@ def predict_segments(class_probabilities: np.ndarray,
         if not probs_are_labels:  # class_probabilities is [T, nb_classes]
             nb_classes = class_probabilities.shape[1]
         else:  # class_probabilities is [T,] with integer entries as class labels
-            nb_classes = int(np.max(class_probabilities))
+            nb_classes = int(da.max(class_probabilities).compute())
         segment_dims = list(range(nb_classes))
 
     if segment_names is None:
-        segment_names = segment_dims
+        segment_names = [str(sd) for sd in segment_dims]
 
-    segments = dict()
+    segments: Dict[str, Any] = dict()
     if len(segment_dims):
         segments['samplerate_Hz'] = samplerate
         segments['index'] = segment_dims
         segments['names'] = segment_names
         if not probs_are_labels:
-            prob = class_probabilities[:, segment_dims]
-            segments['probabilities'] = prob
-            labels = labels_from_probabilities(prob, segment_thres)
+            segments['probabilities'] = class_probabilities[:, segment_dims]
+            labels = da.map_blocks(labels_from_probabilities,
+                                   class_probabilities,
+                                   segment_thres,
+                                   segment_dims,
+                                   dtype=np.intp,
+                                   drop_axis=1)
         else:
             segments['probabilities'] = None
             labels = class_probabilities
 
         # turn into song (0), no song (1) sequence to detect onsets (0->1) and offsets (1->0)
-        song_pred = (labels > 0).astype(np.float)
+        song_binary = (labels > 0).astype(np.int8)
         if segment_fillgap is not None:
-            song_pred = segment_utils.fill_gaps(song_pred, segment_fillgap * samplerate)
+            song_binary = da.map_overlap(segment_utils.fill_gaps,
+                                         song_binary,
+                                         gap_dur=int(segment_fillgap * samplerate),
+                                         depth=(int(segment_fillgap * samplerate + 1), 0),
+                                         boundary="none",
+                                         trim=True,
+                                         align_arrays=True)
         if segment_minlen is not None:
-            song_pred = segment_utils.remove_short(song_pred, segment_minlen * samplerate)
+            song_binary = da.map_overlap(segment_utils.remove_short,
+                                         song_binary,
+                                         min_len=int(segment_minlen * samplerate),
+                                         depth=(int(segment_minlen * samplerate + 1), 0),
+                                         boundary="none",
+                                         trim=True,
+                                         align_arrays=True)
 
         # detect syllable on- and offsets
         # pre- and post-pend 0 so we detect on and offsets at boundaries
-        segments['onsets_seconds'] = np.where(np.diff(np.insert(song_pred, 0, values=[0], axis=0)) == 1)[0].astype(np.float) / samplerate
-        segments['offsets_seconds'] = np.where(np.diff(np.append(song_pred, values=[0], axis=0)) == -1)[0].astype(np.float) / samplerate
-        segments['durations_seconds'] = segments['offsets_seconds'] - segments['onsets_seconds']
+        onsets = da.where(da.diff(song_binary, prepend=0) == 1)[0]
+        offsets = da.where(da.diff(song_binary, append=0) == -1)[0]
+        logging.info("   Detecting syllable on and offsets:")
+        with ProgressBar(minimum=5):
+            onsets, offsets = da.compute(onsets, offsets)
+        segments['onsets_seconds'] = onsets.astype(float) / samplerate
+        segments['offsets_seconds'] = offsets.astype(float) / samplerate
 
-        if len(segment_dims) == 2:  # there is just a single segment type plus noise - in that case we use the gap-filled, short-deleted pred
-            labels = song_pred
-            segments['sequence'] = [str(segment_names[1])] * len(segments['offsets_seconds'])   # syllable-type for each syllable as int
-        elif len(segment_dims) > 2 and segment_labels_by_majority:  # if >1 segment type (plus noise) label sylls by majority vote on un-smoothed labels
+        # there is just a single segment type plus noise - in that case we use the gap-filled, short-deleted pred
+        if len(segment_dims) == 2:
+            labels = song_binary
+            # syllable-type for each syllable as int
+            segments['sequence'] = [str(segment_names[1])] * len(segments['offsets_seconds'])
+        # if >1 segment type (plus noise) label sylls by majority vote on un-smoothed labels
+        elif len(segment_dims) > 2 and segment_labels_by_majority:
             # if no refs provided, use use on/offsets from smoothed labels
             if segment_ref_onsets is None:
                 segment_ref_onsets = segments['onsets_seconds']
             if segment_ref_offsets is None:
                 segment_ref_offsets = segments['offsets_seconds']
 
-            sequence, labels = segment_utils.label_syllables_by_majority(labels,
-                                                                         segment_ref_onsets,
-                                                                         segment_ref_offsets,
+            logging.info("   Labeling by majority:")
+            if len(segment_dims) < np.iinfo('uint8').max:
+                cast_to = np.uint8
+            elif len(segment_dims) < np.iinfo('uint16').max:
+                cast_to = np.uint16
+            elif len(segment_dims) < np.iinfo('uint32').max:
+                cast_to = np.uint32
+            else:
+                cast_to = None
+
+            if cast_to is not None:
+                logging.info(f"   Casting labels to {cast_to}:")
+                labels = da.map_blocks(lambda x: x.astype(cast_to), labels, dtype=np.int16)
+
+            with ProgressBar(minimum=5):
+                labels = labels.compute()
+
+            sequence, labels = segment_utils.label_syllables_by_majority(labels, segment_ref_onsets, segment_ref_offsets,
                                                                          samplerate)
-            segments['sequence'] = sequence  # syllable-type for each syllable as int
+            # syllable-type for each syllable as int
+            segments['sequence'] = sequence
         else:
-            segments['sequence'] = []
+            segments['sequence'] = None
         segments['samples'] = labels
     return segments
 
 
-def predict_events(class_probabilities, samplerate: float = 1.0,
-                   event_dims: List[int] = None, event_names: List[str] = None,
-                   event_thres: float = 0.5, events_offset: float = 0.0, event_dist: float = 100.0,
-                   event_dist_min: float = 0.0, event_dist_max: float = np.inf):
+def _detect_events_oom(event_probability: np.ndarray,
+                       thres: float = 0.70,
+                       min_dist: int = 100,
+                       index: int = 0,
+                       block_info=None,
+                       pad=0) -> np.ndarray:
+    """Wrapper around detect_events that returns 2D np.ndarray for use in predict_oom."""
+    event_indices, event_confidence = event_utils.detect_events(event_probability, thres, min_dist, index)
+
+    if block_info is not None:
+        # add offset from position of chunk in
+        chunk_start_index = block_info[None]['array-location'][0][0]
+        chunk_end_index = block_info[None]['array-location'][0][1]
+
+        event_indices += chunk_start_index
+        # correct for overlap in all but the first chunk
+        event_indices -= pad
+
+        # delete detections in overlapping samples
+        good_events = np.logical_and(event_indices >= chunk_start_index, event_indices < chunk_end_index)
+        if len(good_events):
+            event_confidence = event_confidence[good_events]
+            event_indices = event_indices[good_events]
+        else:
+            event_confidence = []
+            event_indices = []
+    return np.stack((event_indices, event_confidence), axis=1)
+
+
+def predict_events(class_probabilities: np.ndarray,
+                   samplerate: float = 1.0,
+                   event_dims: Sequence[int] = None,
+                   event_names: Sequence[str] = None,
+                   event_thres: float = 0.5,
+                   events_offset: float = 0,
+                   event_dist: float = 100,
+                   event_dist_min: float = 0,
+                   event_dist_max: float = np.inf) -> Dict[str, Any]:
     """[summary]
 
     Args:
-        class_probabilities ([type]): [samples, classes][description]
+        class_probabilities (np.ndarray): [samples, classes][description]
         samplerate (float, optional): Hz
-        event_dims (List[int], optional): [description]. Defaults to np.arange(1, nb_classes) (excludes noise-class at index 0).
+        event_dims (List[int], optional): [description]. Defaults to np.arange(1, nb_classes).
         event_names ([type], optional): [description]. Defaults to event_dims.
         event_thres (float, optional): [description]. Defaults to 0.5.
         events_offset (float, optional): . Defaults to 0 seconds.
@@ -181,17 +298,17 @@ def predict_events(class_probabilities, samplerate: float = 1.0,
         ValueError: [description]
 
     Returns:
-        dict[index/names/sequence/seconds/probabilities]
+        Dict[str, Any]
     """
     if event_dims is None:
         nb_classes = class_probabilities.shape[1]
-        event_dims = np.arange(1, nb_classes).astype(np.uintp)
+        event_dims = np.arange(1, nb_classes)
 
     if event_names is None:
-        event_names = [str(dim) for dim in event_dims]
+        event_names = event_dims
 
     if event_dist_max is None:
-        event_dist_max = np.inf
+        event_dist_max = np.inf  #class_probabilities.shape[0] + 1
 
     events: Dict[str, Any] = dict()
     if len(event_dims):
@@ -203,15 +320,28 @@ def predict_events(class_probabilities, samplerate: float = 1.0,
         events['probabilities'] = []
         events['sequence'] = []
 
+        pad = int(event_dist * samplerate + 1)
         for event_dim, event_name in zip(event_dims, event_names):
-            event_indices, event_probabilities = event_utils.detect_events(class_probabilities, index=event_dim,
-                                                                           thres=event_thres, min_dist=event_dist * samplerate)
-            events_seconds = event_indices.astype(np.float) / samplerate
+
+            event_indices_and_probs = da.map_overlap(_detect_events_oom,
+                                                     class_probabilities,
+                                                     thres=event_thres,
+                                                     min_dist=event_dist * samplerate,
+                                                     index=event_dim,
+                                                     pad=pad,
+                                                     depth=(pad, 0),
+                                                     boundary="none",
+                                                     trim=False,
+                                                     dtype=int,
+                                                     meta=np.array(()))
+            with ProgressBar(minimum=5):
+                event_indices_and_probs = da.compute(event_indices_and_probs)
+            event_indices, event_probabilities = event_indices_and_probs[0][:, 0], event_indices_and_probs[0][:, 1]
+
+            events_seconds = event_indices.astype(float) / samplerate
             events_seconds += events_offset
 
-            good_event_indices = event_utils.event_interval_filter(events_seconds,
-                                                                   event_dist_min, event_dist_max)
-            # good_event_indices = np.arange(len(events_seconds))
+            good_event_indices = event_utils.event_interval_filter(events_seconds, event_dist_min, event_dist_max)
             events['seconds'].extend(events_seconds[good_event_indices])
             events['probabilities'].extend(event_probabilities[good_event_indices])
             events['sequence'].extend([event_name for _ in events_seconds[good_event_indices]])
@@ -219,41 +349,49 @@ def predict_events(class_probabilities, samplerate: float = 1.0,
     return events
 
 
-def predict_song(class_probabilities: np.ndarray, params: Dict[str, Any],
-                 event_thres: float = 0.5, event_dist: float = 0.01,
-                 event_dist_min: float = 0, event_dist_max: float = None,
+def predict_song(class_probabilities: np.ndarray,
+                 params: Dict[str, Any],
+                 event_thres: float = 0.5,
+                 event_dist: float = 0.01,
+                 event_dist_min: float = 0,
+                 event_dist_max: float = np.inf,
                  segment_ref_onsets: Optional[List[float]] = None,
                  segment_ref_offsets: Optional[List[float]] = None,
-                 segment_thres: float = 0.5, segment_minlen: float = None,
+                 segment_thres: float = 0.5,
+                 segment_minlen: float = None,
                  segment_fillgap: float = None):
 
     samplerate = params['samplerate_x_Hz']
     events_offset = 0
 
     segment_dims = np.where([val == 'segment' for val in params['class_types']])[0]
-    segment_names = [params['class_names'][segment_dim] for segment_dim in segment_dims]
-    segments = predict_segments(class_probabilities, samplerate,
-                                segment_dims, segment_names,
-                                segment_ref_onsets, segment_ref_offsets,
-                                segment_thres, segment_minlen, segment_fillgap)
+    segment_names = [str(params['class_names'][segment_dim]) for segment_dim in segment_dims]
+    segments = predict_segments(class_probabilities, samplerate, segment_dims, segment_names, segment_ref_onsets,
+                                segment_ref_offsets, segment_thres, segment_minlen, segment_fillgap)
 
     event_dims = np.where([val == 'event' for val in params['class_types']])[0]
-    event_names = [params['class_names'][event_dim] for event_dim in event_dims]
-    events = predict_events(class_probabilities, samplerate,
-                            event_dims, event_names,
-                            event_thres, events_offset, event_dist,
+    event_names = [str(params['class_names'][event_dim]) for event_dim in event_dims]
+    events = predict_events(class_probabilities, samplerate, event_dims, event_names, event_thres, events_offset, event_dist,
                             event_dist_min, event_dist_max)
     return events, segments
 
 
-def predict(x: np.ndarray, model_save_name: str = None, verbose: int = 1,
+def predict(x: np.ndarray,
+            model_save_name: str = None,
+            verbose: int = 1,
             batch_size: int = None,
-            model: models.keras.models.Model = None, params: dict = None,
-            event_thres: float = 0.5, event_dist: float = 0.01,
-            event_dist_min: float = 0, event_dist_max: float = None,
-            segment_thres: float = 0.5, segment_use_optimized: bool = True,
-            segment_minlen: float = None, segment_fillgap: float = None,
-            pad: bool = True, prepend_data_padding: bool = True):
+            model: models.keras.models.Model = None,
+            params: Dict = None,
+            event_thres: float = 0.5,
+            event_dist: float = 0.01,
+            event_dist_min: float = 0,
+            event_dist_max: float = None,
+            segment_thres: float = 0.5,
+            segment_use_optimized: bool = True,
+            segment_minlen: float = None,
+            segment_fillgap: float = None,
+            pad: bool = True,
+            prepend_data_padding: bool = True):
     """[summary]
 
     Usage:
@@ -340,7 +478,7 @@ def predict(x: np.ndarray, model_save_name: str = None, verbose: int = 1,
         # pad with end val to fill
         x = np.pad(x, ((0, pad_len), (0, 0)), mode='edge')
 
-    class_probabilities = predict_probabililties(x, model, params, verbose, prepend_data_padding)
+    class_probabilities = predict_probabilities(x, model, params, verbose, prepend_data_padding)
 
     if pad:
         # trim probs to original len of x
@@ -350,23 +488,36 @@ def predict(x: np.ndarray, model_save_name: str = None, verbose: int = 1,
         # class_probabilities[-pad_len:, 1:] = 0
         # class_probabilities[-pad_len:, 0] = 1
 
-    events, segments = predict_song(class_probabilities=class_probabilities, params=params,
-                                    event_thres=event_thres, event_dist=event_dist,
-                                    event_dist_min=event_dist_min, event_dist_max=event_dist_max,
-                                    segment_ref_onsets=None, segment_ref_offsets=None,
-                                    segment_thres=segment_thres, segment_minlen=segment_minlen,
+    events, segments = predict_song(class_probabilities=class_probabilities,
+                                    params=params,
+                                    event_thres=event_thres,
+                                    event_dist=event_dist,
+                                    event_dist_min=event_dist_min,
+                                    event_dist_max=event_dist_max,
+                                    segment_ref_onsets=None,
+                                    segment_ref_offsets=None,
+                                    segment_thres=segment_thres,
+                                    segment_minlen=segment_minlen,
                                     segment_fillgap=segment_fillgap)
 
     return events, segments, class_probabilities, params['class_names']
 
 
-def cli_predict(path: str, model_save_name: str, *,
-                save_filename: Optional[str] = None, save_format: str = 'csv',
-                verbose: int = 1, batch_size: Optional[int] = None,
-                event_thres: float = 0.5, event_dist: float = 0.01,
-                event_dist_min: float = 0, event_dist_max: Optional[float] = None,
-                segment_thres: float = 0.5, segment_use_optimized: Optional[bool] = None,
-                segment_minlen: Optional[float] = None, segment_fillgap: Optional[float] = None,
+def cli_predict(path: str,
+                model_save_name: str,
+                *,
+                save_filename: Optional[str] = None,
+                save_format: str = 'csv',
+                verbose: int = 1,
+                batch_size: Optional[int] = None,
+                event_thres: float = 0.5,
+                event_dist: float = 0.01,
+                event_dist_min: float = 0,
+                event_dist_max: Optional[float] = None,
+                segment_thres: float = 0.5,
+                segment_use_optimized: Optional[bool] = None,
+                segment_minlen: Optional[float] = None,
+                segment_fillgap: Optional[float] = None,
                 resample: bool = True):
     """Predict song labels for a wav file or a folder of wav files.
 
@@ -432,9 +583,10 @@ def cli_predict(path: str, model_save_name: str, *,
         logging.info(f"   Loading data from {recording_filename}.")
         try:
             # else if path is file - predict only on file but make it single-item list
-            x, fs_audio = librosa.load(recording_filename, sr=None, mono=False)
-            x = x.T
-            x = x[:, np.newaxis] if x.ndim == 1 else x  # adds singleton dim for single-channel wavs
+            fs_audio, x = scipy.io.wavfile.read(recording_filename)
+
+            if x.ndim == 1:
+                x = x[:, np.newaxis]
 
             if resample and fs_audio != fs_model:
                 logging.info(f"   Resampling. Audio rate is {fs_audio}Hz but model was trained on data with {fs_model}Hz.")
@@ -442,22 +594,25 @@ def cli_predict(path: str, model_save_name: str, *,
 
             logging.info(f"   Annotating using model at {model_save_name}.")
             # TODO: load model once, provide as direct arg
-            events, segments, class_probabilities, class_names = predict(x, None, verbose, batch_size,
-                                                                         model, params,
-                                                                         event_thres, event_dist, event_dist_min, event_dist_max,
-                                                                         segment_thres, segment_use_optimized, segment_minlen, segment_fillgap)
+            events, segments, class_probabilities, class_names = predict(x, None, verbose, batch_size, model, params,
+                                                                         event_thres, event_dist, event_dist_min,
+                                                                         event_dist_max, segment_thres, segment_use_optimized,
+                                                                         segment_minlen, segment_fillgap)
 
             if 'event' in params["class_types"]:
                 logging.info(f"   found {len(events['seconds'])} instances of events '{list(set(events['sequence']))}'.")
             if 'segment' in params["class_types"]:
-                logging.info(f"   found {len(segments['onsets_seconds'])} instances of segments '{list(set(segments['sequence']))}'.")
+                logging.info(
+                    f"   found {len(segments['onsets_seconds'])} instances of segments '{list(set(segments['sequence']))}'.")
 
             if save_format == 'h5':
                 # turn events and segments into df!
-                d = {'events': events,
-                     'segments': segments,
-                     'class_probabilities': class_probabilities,
-                     'class_names': class_names}
+                d = {
+                    'events': events,
+                    'segments': segments,
+                    'class_probabilities': class_probabilities,
+                    'class_names': class_names
+                }
                 if save_filename is None:
                     save_filename = os.path.splitext(recording_filename)[0] + '_das.h5'
                 logging.info(f"   Saving results to {save_filename}.")
