@@ -7,14 +7,10 @@ from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint, ReduceLRO
 from tensorflow import keras
 import sklearn.utils
 import os
-import rich.logging
 import yaml
-from typing import List, Optional, Tuple, Dict, Any
+from typing import List, Optional, Tuple, Dict, Any, Union
 from . import data, models, utils, predict, io, evaluate, tracking, data_hash, augmentation, postprocessing  #, timeseries
 
-FORMAT = "%(message)s"
-logging.basicConfig(level="INFO", format=FORMAT, datefmt="[%X]", handlers=[rich.logging.RichHandler()])
-
 logger = logging.getLogger(__name__)
 
 try:  # fixes cuDNN error when using LSTM layer
@@ -38,7 +34,7 @@ def train(*,
           nb_filters: int = 16,
           kernel_size: int = 16,
           nb_conv: int = 3,
-          use_separable: List[bool] = [False],
+          use_separable: List[bool] = False,
           nb_hist: int = 1024,
           ignore_boundaries: bool = True,
           batch_norm: bool = True,
@@ -74,12 +70,12 @@ def train(*,
           balance: bool = False,
           version_data: bool = True,
           post_opt: bool = False,
-          fill_gaps_min: Optional[float] = None,
-          fill_gaps_max: Optional[float] = None,
-          fill_gaps_steps: Optional[float] = None,
-          min_len_min: Optional[float] = None,
-          min_len_max: Optional[float] = None,
-          min_len_steps: Optional[float] = None,
+          fill_gaps_min: float = 0.0005,
+          fill_gaps_max: float = 1.0,
+          fill_gaps_steps: int = 20,
+          min_len_min: float = 0.0005,
+          min_len_max: float = 1.0,
+          min_len_steps: int = 20,
           resnet_compute: bool = False,
           resnet_train: bool = False,
           _qt_progress: bool = False) -> Tuple[keras.Model, Dict[str, Any]]:
@@ -107,7 +103,6 @@ def train(*,
                            or with SAVE_DIR/SAVE_NAME if SAVE_PREFIX is empty.
                            Defaults to the timestamp YYYYMMDD_hhmmss.
         model_name (str): Network architecture to use.
-                          Use `tcn` (TCN) or `tcn_stft` (TCN with STFT frontend).
                           See das.models for a description of all models.
                           Defaults to `tcn`.
         nb_filters (int): Number of filters per layer.
@@ -197,12 +192,12 @@ def train(*,
 
         post_opt (bool): Optimize post processing (delete short detections, fill brief gaps).
                         Defaults to False.
-        fill_gaps_min (Optional[float]): None
-        fill_gaps_max (Optional[float]): None
-        fill_gaps_steps (Optional[float]): None
-        min_len_min (Optional[float]): None
-        min_len_max (Optional[float]): None
-        min_len_steps (Optional[float]): None
+        fill_gaps_min (float): Defaults to 0.0005 seconds.
+        fill_gaps_max (float): Defaults to 1 second.
+        fill_gaps_steps (int): Defaults to 20.
+        min_len_min (float): Defaults to 0.0005 seconds.
+        min_len_max (float): Defaults to 1 second.
+        min_len_steps (int): Defaults to 20.
 
         resnet_compute (bool): Defaults to False.
         resnet_train (bool): Defaults to False.
@@ -360,7 +355,7 @@ def train(*,
     logger.info(f"Validation data:")
     logger.info(f"   {val_gen}")
 
-    logger.info('building network')
+    logger.info('Building network')
     try:
         model = models.model_dict[model_name](**params)
     except KeyError as e:
@@ -415,13 +410,8 @@ def train(*,
     if post_opt:
         logger.info('OPTIMIZING POSTPROCESSING:')
 
-        gap_durs = None
-        if fill_gaps_min is not None and fill_gaps_max is not None and fill_gaps_steps is not None:
-            gap_durs = np.geomspace(fill_gaps_min, fill_gaps_max, fill_gaps_steps)
-
-        min_lens = None
-        if min_len_min is not None and min_len_max is not None and min_len_steps is not None:
-            min_lens = np.geomspace(min_len_min, min_len_max, min_len_steps)
+        gap_durs = np.geomspace(fill_gaps_min, fill_gaps_max, fill_gaps_steps)
+        min_lens = np.geomspace(min_len_min, min_len_max, min_len_steps)
 
         best_gap_dur, best_min_len, scores = postprocessing.optimize(dataset_path=data_dir,
                                                                      model_save_name=save_name,
@@ -449,23 +439,14 @@ def train(*,
     # TEST
     # TODO use postprocessing params
     logger.info('TESTING:')
-    if len(d['test']['x']) < nb_hist:
+    if 'test' not in d or len(d['test']['x']) < nb_hist:
         logger.info('   No test data - skipping final evaluation step.')
     else:
         logger.info(f'   Re-loading last best model from {checkpoint_save_name}.')
         model.load_weights(checkpoint_save_name)
 
         logger.info('   Predicting.')
-        x_name = 'x'
-        if x_suffix:
-            x_name += f'_{x_suffix}'
-        y_name = 'y'
-        if y_suffix:
-            y_name += f'_{y_suffix}'
-        x_test, y_test, y_pred = evaluate.evaluate_probabilities(x=d['test'][x_name],
-                                                                 y=d['test'][y_name],
-                                                                 model=model,
-                                                                 params=params)
+        x_test, y_test, y_pred = evaluate.evaluate_probabilities(x=d['test']['x'], y=d['test']['y'], model=model, params=params)
 
         labels_test = predict.labels_from_probabilities(y_test)
         labels_pred = predict.labels_from_probabilities(y_pred)
